@@ -107,6 +107,7 @@ class ClientWindowBase(ClientWidgetBase):
         #initialize gravity early:
         sc = typedict(metadata.dictget("size-constraints", {}))
         self.window_gravity = OVERRIDE_GRAVITY or sc.intget("gravity", DEFAULT_GRAVITY)
+        self.set_decorated(metadata.boolget("decorations", True))
 
 
     def get_desktop_workspace(self):
@@ -251,7 +252,7 @@ class ClientWindowBase(ClientWidgetBase):
             except Exception as e:
                 log.error("Error parsing window title:")
                 log.error(" %s", e)
-                title = b""
+                title = ""
             self.set_title(title)
 
         if b"icon-title" in metadata:
@@ -339,8 +340,14 @@ class ClientWindowBase(ClientWidgetBase):
                     self.deiconify()
 
         if b"decorations" in metadata:
-            self.set_decorated(metadata.boolget("decorations"))
-            self.apply_geometry_hints(self.geometry_hints)
+            decorated = metadata.boolget("decorations", True)
+            was_decorated = self.get_decorated()
+            if WIN32 and decorated!=was_decorated:
+                log.info("decorations flag toggled, now %s, re-initializing window", decorated)
+                self.idle_add(self._client.reinit_window, self._id, self)
+            else:
+                self.set_decorated(metadata.boolget("decorations"))
+                self.apply_geometry_hints(self.geometry_hints)
 
         if b"above" in metadata:
             above = metadata.boolget("above")
@@ -444,6 +451,9 @@ class ClientWindowBase(ClientWidgetBase):
             geomlog("intpair(%s)=%s", a, v)
             if v:
                 v1, v2 = v
+                if a==b"maximum-size" and v1>=32000 and v2>=32000 and WIN32:
+                    #causes problems, see #2714
+                    continue
                 sv1 = client.sx(v1)
                 sv2 = client.sy(v2)
                 if a in (b"base-size", b"increment"):
