@@ -484,10 +484,10 @@ def do_parse_cmdline(cmdline, defaults):
                           dest="displayfd", default=defaults.displayfd,
                           help="The xpra server will write the display number back on this file descriptor"
                           +" as a newline-terminated string.")
-        group.add_option("--fake-xinerama", action="store", metavar="yes|no",
+        group.add_option("--fake-xinerama", action="store", metavar="path|auto|no",
                           dest="fake_xinerama",
                           default=defaults.fake_xinerama,
-                          help="Setup fake xinerama support for the session."
+                          help="Setup fake xinerama support for the session. You can specify the path to the libfakeXinerama.so library or a boolean."
                           +" Default: %s." % enabled_str(defaults.fake_xinerama))
     else:
         ignore({
@@ -998,10 +998,10 @@ def do_parse_cmdline(cmdline, defaults):
                       +" can be specified multiple times."
                       +" Default: %s." % csv(
                           ("'%s'" % x) for x in (defaults.env or []) if not x.startswith("#")))
-    group.add_option("--challenge-handlers", action="store",
-                      dest="challenge_handlers", default=defaults.challenge_handlers,
+    group.add_option("--challenge-handlers", action="append",
+                      dest="challenge_handlers", default=[],
                       help="Which handlers to use for processing server authentication challenges."
-                      +" Default: %default.")
+                      +" Default: %s." % csv(defaults.challenge_handlers))
     group.add_option("--password-file", action="append",
                       dest="password_file", default=defaults.password_file,
                       help="The file containing the password required to connect"
@@ -1039,8 +1039,8 @@ def do_parse_cmdline(cmdline, defaults):
                       help="How often to synchronize the virtual framebuffer used for X11 seamless servers "
                       +"(0 to disable)."
                       +" Default: %s." % defaults.sync_xvfb)
-    group.add_option("--socket-dirs", action="append",
-                      dest="socket_dirs", default=[],
+    group.add_option("--socket-dirs", action="store",
+                      dest="socket_dirs", default=defaults.socket_dirs,
                       help="Directories to look for the socket files in."
                       +" Default: %s." % os.path.pathsep.join("'%s'" % x for x in defaults.socket_dirs))
     default_socket_dir_str = defaults.socket_dir or "$XPRA_SOCKET_DIR or the first valid directory in socket-dirs"
@@ -1217,6 +1217,10 @@ def do_parse_cmdline(cmdline, defaults):
         #use the default:
         options.bind = defaults_bind
 
+    #only use the default challenge-handlers if the user hasn't specified any:
+    if not options.challenge_handlers:
+        options.challenge_handlers = defaults.challenge_handlers
+
     #only use the default key-shortcut list if the user hasn't specified one:
     if not options.key_shortcut:
         options.key_shortcut = defaults.key_shortcut
@@ -1249,7 +1253,7 @@ def do_parse_cmdline(cmdline, defaults):
 
     #special case for things stored as lists, but command line option is a CSV string:
     #and may have "none" or "all" special values
-    fixup_options(options, defaults)
+    fixup_options(options)
 
     try:
         options.dpi = int(options.dpi)
@@ -1297,8 +1301,6 @@ def do_validate_encryption(auth, tcp_auth,
         return
     from xpra.net.crypto import crypto_backend_init
     crypto_backend_init()
-    if not encryption and not tcp_encryption:
-        return
     env_key = os.environ.get("XPRA_ENCRYPTION_KEY")
     pass_key = os.environ.get("XPRA_PASSWORD")
     from xpra.net.crypto import ENCRYPTION_CIPHERS
@@ -1314,8 +1316,8 @@ def do_validate_encryption(auth, tcp_auth,
         raise InitException("encryption %s cannot be used without an authentication module or keyfile"
                             +" (see --encryption-keyfile option)" % encryption)
     if tcp_encryption and not tcp_encryption_keyfile and not env_key and not tcp_auth:
-        raise InitException("tcp-encryption %s cannot be used without a tcp authentication module or keyfile "
-                            +" (see --tcp-encryption-keyfile option)" % tcp_encryption)
+        raise InitException("tcp-encryption %s cannot be used without a tcp authentication module or keyfile " % tcp_encryption
+                            +" (see --tcp-encryption-keyfile option)")
     if pass_key and env_key and pass_key==env_key:
         raise InitException("encryption and authentication should not use the same value")
     #discouraged but not illegal:
