@@ -13,6 +13,7 @@ import socket
 import signal
 import threading
 import traceback
+from collections import OrderedDict
 from weakref import WeakKeyDictionary
 from time import sleep, time
 
@@ -21,7 +22,13 @@ from xpra.version_util import (
     get_platform_info, get_host_info,
     )
 from xpra.scripts.server import deadly_signal
+from xpra.server.server_util import write_pidfile, rm_pidfile
 from xpra.scripts.config import InitException, parse_bool, python_platform, parse_with_unit, FALSE_OPTIONS, TRUE_OPTIONS
+from xpra.net.common import may_log_packet, SOCKET_TYPES
+from xpra.net.socket_util import (
+    hosts, mdns_publish, peek_connection,
+    add_listen_socket, accept_connection,
+    )
 from xpra.net.bytestreams import (
     SocketConnection, SSLSocketConnection,
     log_new_connection, pretty_socket, SOCKET_TIMEOUT,
@@ -32,14 +39,15 @@ from xpra.net.digest import get_salt, gendigest, choose_digest
 from xpra.platform import set_name
 from xpra.platform.paths import get_app_dir
 from xpra.os_util import (
-    filedata_nocrlf, get_machine_id, get_user_uuid, platform_name,
+    filedata_nocrlf, get_machine_id, get_user_uuid, platform_name, get_ssh_port,
     strtobytes, bytestostr, get_hex_uuid,
-    getuid, monotonic_time, get_peercred, hexstr,
+    getuid, monotonic_time, hexstr,
     WIN32, POSIX, PYTHON3, BITS,
     )
-from xpra.server.background_worker import stop_worker, get_worker
+from xpra.server.background_worker import stop_worker, get_worker, add_work_item
 from xpra.make_thread import start_thread
 from xpra.util import (
+    first_time,
     csv, merge_dicts, typedict, notypedict, flatten_dict, parse_simple_dict,
     repr_ellipsized, dump_all_frames, nonl, envint, envbool, envfloat,
     SERVER_SHUTDOWN, SERVER_UPGRADE, LOGIN_TIMEOUT, DONE, PROTOCOL_ERROR,
@@ -57,6 +65,7 @@ commandlog = Logger("command")
 authlog = Logger("auth")
 timeoutlog = Logger("timeout")
 dbuslog = Logger("dbus")
+mdnslog = Logger("mdns")
 
 main_thread = threading.current_thread()
 
@@ -64,7 +73,6 @@ MAX_CONCURRENT_CONNECTIONS = envint("XPRA_MAX_CONCURRENT_CONNECTIONS", 100)
 SIMULATE_SERVER_HELLO_ERROR = envbool("XPRA_SIMULATE_SERVER_HELLO_ERROR", False)
 SERVER_SOCKET_TIMEOUT = envfloat("XPRA_SERVER_SOCKET_TIMEOUT", "0.1")
 LEGACY_SALT_DIGEST = envbool("XPRA_LEGACY_SALT_DIGEST", True)
-PEEK_TIMEOUT = envint("XPRA_PEEK_TIMEOUT", 1)
 CHALLENGE_TIMEOUT = envint("XPRA_CHALLENGE_TIMEOUT", 120)
 
 
@@ -173,9 +181,6 @@ def get_thread_info(proto=None, protocols=()):
         log.error("failed to get frame info: %s", e)
     return info
 
-def notimplemented(*_args):
-    raise NotImplementedError()
-
 
 class ServerCore(object):
     """
@@ -192,6 +197,7 @@ class ServerCore(object):
         self.child_reaper = None
         self.original_desktop_display = None
         self.session_type = "unknown"
+        self.display_name = ""
 
         self._closing = False
         self._upgrading = False
@@ -203,31 +209,36 @@ class ServerCore(object):
         self._tcp_proxy_clients = []
         self._tcp_proxy = ""
         self._rfb_upgrade = 0
-        self._ssl_wrap_socket = notimplemented
+        self._ssl_wrap_socket = None
         self._accept_timeout = SOCKET_TIMEOUT + 1
         self.ssl_mode = None
         self._html = False
         self._www_dir = None
         self._http_headers_dir = None
         self._aliases = {}
-        self._reverse_aliases = {}
-        self.socket_types = {}
         self.socket_info = {}
+        self.socket_cleanup = []
         self.socket_verify_timer = WeakKeyDictionary()
         self.socket_rfb_upgrade_timer = WeakKeyDictionary()
         self._max_connections = MAX_CONCURRENT_CONNECTIONS
         self._socket_timeout = SERVER_SOCKET_TIMEOUT
         self._ws_timeout = 5
         self._socket_dir = None
+        self.dbus_pid = 0
+        self.dbus_env = {}
         self.dbus_control = False
         self.dbus_server = None
         self.unix_socket_paths = []
         self.touch_timer = None
         self.exec_cwd = os.getcwd()
+        self.pidfile = None
+        self.pidinode = 0
 
         self.session_name = u""
 
         #Features:
+        self.mdns = False
+        self.mdns_publishers = {}
         self.encryption = None
         self.encryption_keyfile = None
         self.tcp_encryption = None
@@ -256,6 +267,7 @@ class ServerCore(object):
         raise NotImplementedError()
 
     def init_when_ready(self, callbacks):
+        log("init_when_ready(%s)", callbacks)
         self._when_ready = callbacks
 
 
@@ -281,11 +293,49 @@ class ServerCore(object):
         self.exit_with_client = opts.exit_with_client
         self.server_idle_timeout = opts.server_idle_timeout
         self.readonly = opts.readonly
-        self.ssl_mode = opts.ssl
         self.ssh_upgrade = opts.ssh_upgrade
         self.dbus_control = opts.dbus_control
+        self.pidfile = opts.pidfile
+        self.mdns = opts.mdns
         self.init_html_proxy(opts)
         self.init_auth(opts)
+        self.init_ssl(opts)
+        if self.pidfile:
+            self.pidinode = write_pidfile(self.pidfile)
+
+
+    def init_ssl(self, opts):
+        self.ssl_mode = opts.ssl
+        if self.ssl_mode.lower() in FALSE_OPTIONS:
+            return
+        need_ssl = False
+        if self.ssl_mode in TRUE_OPTIONS or opts.bind_ssl or opts.bind_wss:
+            need_ssl = True
+        elif opts.bind_tcp or opts.bind_ws:
+            if self.ssl_mode=="auto" and opts.ssl_cert:
+                need_ssl = True
+            elif self.ssl_mode=="tcp" and opts.bind_tcp:
+                need_ssl = True
+            elif self.ssl_mode=="www":
+                need_ssl = True
+        if not need_ssl:
+            return
+        from xpra.scripts.main import ssl_wrap_socket_fn
+        try:
+            self._ssl_wrap_socket = ssl_wrap_socket_fn(opts, server_side=True)
+            log("init_ssl() wrap_socket_fn=%s", self._ssl_wrap_socket)
+        except Exception as e:
+            log("SSL error", exc_info=True)
+            cpaths = csv("'%s'" % x for x in (opts.ssl_cert, opts.ssl_key) if x)
+            raise InitException("cannot create SSL sockets, check your certificate paths (%s): %s" % (cpaths, e))
+
+    def server_ready(self):
+        return True
+
+    def server_init(self):
+        if self.mdns:
+            add_work_item(self.mdns_publish)
+        self.start_listen_sockets()
 
     def setup(self):
         self.init_packet_handlers()
@@ -309,14 +359,24 @@ class ServerCore(object):
         self._closing = True
         self.cleanup()
         def quit_timer():
-            log.debug("quit_timer()")
-            stop_worker(True)
+            w = get_worker()
+            log("quit_timer() worker=%s", w)
+            if w and w.is_alive():
+                #wait up to 1 second for the worker thread to exit
+                try:
+                    w.wait(1)
+                except:
+                    pass
+                if w.is_alive():
+                    #still alive, force stop:
+                    stop_worker(True)
+                    try:
+                        w.wait(1)
+                    except:
+                        pass
             self.quit(upgrading)
-        #if from a signal, just force quit:
         stop_worker()
-        #not from signal: use force stop worker after delay
-        self.timeout_add(250, stop_worker, True)
-        self.timeout_add(500, quit_timer)
+        self.timeout_add(250, quit_timer)
         def force_quit():
             log.debug("force_quit()")
             from xpra import os_util
@@ -347,8 +407,6 @@ class ServerCore(object):
 
 
     def run(self):
-        self.print_run_info()
-        self.print_screen_info()
         self.install_signal_handlers(self.signal_quit)
         def start_ready_callbacks():
             for x in self._when_ready:
@@ -362,9 +420,10 @@ class ServerCore(object):
         self.idle_add(start_ready_callbacks)
         self.idle_add(self.reset_server_timeout)
         self.idle_add(self.server_is_ready)
+        self.idle_add(self.print_run_info)
         self.do_run()
-        log("run() returning %s", self._upgrading)
-        return self._upgrading
+        log("run()")
+        return 0
 
     def server_is_ready(self):
         log.info("xpra is ready.")
@@ -373,12 +432,14 @@ class ServerCore(object):
     def do_run(self):
         raise NotImplementedError()
 
-    def cleanup(self, *_args):
+    def cleanup(self):
         netlog("cleanup() stopping %s tcp proxy clients: %s", len(self._tcp_proxy_clients), self._tcp_proxy_clients)
         for p in tuple(self._tcp_proxy_clients):
             p.quit()
         netlog("cleanup will disconnect: %s", self._potential_protocols)
         self.cancel_touch_timer()
+        if self.mdns_publishers:
+            add_work_item(self.mdns_cleanup)
         if self._upgrading:
             reason = SERVER_UPGRADE
         else:
@@ -389,16 +450,55 @@ class ServerCore(object):
         self.cleanup_protocols(protocols, reason, True)
         self._potential_protocols = []
         self.cleanup_udp_listeners()
+        self.cleanup_sockets()
         self.cleanup_dbus_server()
+        if not self._upgrading:
+            self.stop_dbus_server()
+        if self.pidfile:
+            self.pidinode = rm_pidfile(self.pidfile, self.pidinode)
 
     def do_cleanup(self):
         #allow just a bit of time for the protocol packet flush
         sleep(0.1)
 
 
+    def cleanup_sockets(self):
+        #stop listening for IO events:
+        for sc in self.socket_cleanup:
+            sc()
+        #actually close the socket:
+        si = self._socket_info
+        self._socket_info = ()
+        for socktype, _, info, cleanup in si:
+            log("cleanup_sockets() calling %s for %s %s", cleanup, socktype, info)
+            try:
+                cleanup()
+            except Exception:
+                log("cleanup error on %s", cleanup, exc_info=True)
+
+
     ######################################################################
     # dbus:
+    def init_dbus(self, dbus_pid, dbus_env):
+        if not POSIX:
+            return
+        self.dbus_pid = dbus_pid
+        self.dbus_env = dbus_env
+
+    def stop_dbus_server(self):
+        dbuslog("stop_dbus_server() dbus_pid=%s", self.dbus_pid)
+        if not self.dbus_pid:
+            return
+        try:
+            os.kill(self.dbus_pid, signal.SIGINT)
+        except Exception as e:
+            dbuslog("os.kill(%i, SIGINT)", self.dbus_pid, exc_info=True)
+            dbuslog.warn("Warning: error trying to stop dbus with pid %i:", self.dbus_pid)
+            dbuslog.warn(" %s", e)
+
     def init_dbus_server(self):
+        if not POSIX:
+            return
         dbuslog("init_dbus_server() dbus_control=%s", self.dbus_control)
         dbuslog("init_dbus_server() env: %s", dict((k,v) for k,v in os.environ.items()
                                                if bytestostr(k).startswith("DBUS_")))
@@ -419,7 +519,7 @@ class ServerCore(object):
             ds.cleanup()
             self.dbus_server = None
 
-    def make_dbus_server(self):
+    def make_dbus_server(self):     #pylint: disable=useless-return
         dbuslog("make_dbus_server() no dbus server for %s", self)
         return None
 
@@ -491,7 +591,7 @@ class ServerCore(object):
             httplog.error(" '%s' does not exist", self._www_dir)
             self._html = False
         if self._html:
-            httplog.info("serving html content from: %s", self._www_dir)
+            httplog.info("serving html content from '%s'", self._www_dir)
         if self._html and self._tcp_proxy:
             httplog.warn("Warning: the built in html server is enabled,")
             httplog.warn(" disabling the tcp-proxy option")
@@ -506,7 +606,7 @@ class ServerCore(object):
             self.auth_classes["named-pipe"] = auth
         else:
             self.auth_classes["unix-domain"] = auth
-        for x in ("tcp", "ws", "wss", "ssl", "ssh", "rfb", "vsock", "udp"):
+        for x in SOCKET_TYPES:
             opts_value = getattr(opts, "%s_auth" % x)
             self.auth_classes[x] = self.get_auth_modules(x, opts_value, opts)
         authlog("init_auth(..) auth=%s", self.auth_classes)
@@ -520,7 +620,13 @@ class ServerCore(object):
     def get_auth_module(self, socket_type, auth_str, opts):
         authlog("get_auth_module(%s, %s, {..})", socket_type, auth_str)
         #separate options from the auth module name
-        parts = auth_str.split(":", 1)
+        #either with ":" or "," as separator
+        scpos = auth_str.find(":")
+        cpos = auth_str.find(",")
+        if cpos<0 or scpos<cpos:
+            parts = auth_str.split(":", 1)
+        else:
+            parts = auth_str.split(",", 1)
         if len(parts)==1 and auth_str.find(",")>0:
             parts = auth_str.split(",", 1)
         auth = parts[0]
@@ -598,11 +704,14 @@ class ServerCore(object):
 
 
     def print_run_info(self):
+        add_work_item(self.do_print_run_info)
+
+    def do_print_run_info(self):
         log.info("xpra %s version %s %i-bit", self.get_server_mode(), full_version_str(), BITS)
         try:
             pinfo = get_platform_info()
             osinfo = " on %s" % platform_name(sys.platform, pinfo.get("linux_distribution") or pinfo.get("sysrelease", ""))
-        except:
+        except Exception:
             log("platform name error:", exc_info=True)
             osinfo = ""
         if POSIX:
@@ -617,6 +726,7 @@ class ServerCore(object):
             except:
                 log.info(" uid=%i, gid=%i", uid, gid)
         log.info(" running with pid %s%s", os.getpid(), osinfo)
+        self.idle_add(self.print_screen_info)
 
     def notify_new_user(self, ss):
         pass
@@ -641,9 +751,110 @@ class ServerCore(object):
     # sockets / connections / packets:
     def init_sockets(self, sockets):
         self._socket_info = sockets
+
+
+    def mdns_publish(self):
+        if not self.mdns:
+            return
+        #find all the records we want to publish:
+        mdns_recs = {}
+        for socktype, _, info, _ in self._socket_info:
+            socktypes = self.get_mdns_socktypes(socktype)
+            mdnslog("mdns_publish() info=%s, socktypes(%s)=%s", info, socktype, socktypes)
+            for st in socktypes:
+                recs = mdns_recs.setdefault(st, [])
+                if socktype=="unix-domain":
+                    assert st=="ssh"
+                    host = "*"
+                    iport = get_ssh_port()
+                    if not iport:
+                        continue
+                else:
+                    host, iport = info
+                for h in hosts(host):
+                    rec = (h, iport)
+                    if rec not in recs:
+                        recs.append(rec)
+                mdnslog("mdns_publish() recs[%s]=%s", st, recs)
+        mdns_info = self.get_mdns_info()
+        self.mdns_publishers = {}
+        for mdns_mode, listen_on in mdns_recs.items():
+            info = OrderedDict(mdns_info)
+            info["mode"] = mdns_mode
+            aps = mdns_publish(self.display_name, listen_on, info)
+            for ap in aps:
+                ap.start()
+                self.mdns_publishers[ap] = mdns_mode
+
+    def get_mdns_socktypes(self, socktype):
+        #for a given socket type,
+        #what socket types we should expose via mdns
+        if socktype in ("vsock", "named-pipe"):
+            #cannot be accessed remotely
+            return ()
+        ssh_access = get_ssh_port()>0   #and opts.ssh.lower().strip() not in FALSE_OPTIONS
+        ssl = bool(self._ssl_wrap_socket)
+        #only available with the RFBServer
+        rfb_upgrades = getattr(self, "_rfb_upgrade", False)
+        socktypes = [socktype]
+        if socktype=="tcp":
+            if ssl:
+                socktypes.append("ssl")
+            if self._html:
+                socktypes.append("ws")
+            if self._html and ssl:
+                socktypes.append("wss")
+            if self.ssh_upgrade:
+                socktypes.append("ssh")
+            if rfb_upgrades:
+                socktypes.append("rfb")
+        elif socktype=="ws":
+            if ssl:
+                socktypes.append("wss")
+        elif socktype=="unix-domain":
+            if ssh_access:
+                socktypes = ["ssh"]
+        return socktypes
+
+    def get_mdns_info(self):
+        from xpra.platform.info import get_username
+        mdns_info = OrderedDict({
+            "display"  : self.display_name,
+            "username" : get_username(),
+            "uuid"     : self.uuid,
+            "platform" : sys.platform,
+            "type"     : self.session_type,
+            })
+        MDNS_EXPOSE_NAME = envbool("XPRA_MDNS_EXPOSE_NAME", True)
+        if MDNS_EXPOSE_NAME and self.session_name:
+            mdns_info["name"] = self.session_name
+        return mdns_info
+
+    def mdns_cleanup(self):
+        mp = self.mdns_publishers
+        self.mdns_publishers = {}
+        for ap in tuple(mp.keys()):
+            ap.stop()
+
+    def mdns_update(self):
+        if not self.mdns:
+            return
+        txt = self.get_mdns_info()
+        for mdns_publisher, mode in dict(self.mdns_publishers).items():
+            info = dict(txt)
+            info["mode"] = mode
+            try:
+                mdns_publisher.update_txt(info)
+            except Exception as e:
+                mdnslog("mdns_update: %s(%s)", mdns_publisher.update_txt, info, exc_info=True)
+                mdnslog.warn("Warning: mdns update failed")
+                mdnslog.warn(" %s", e)
+
+
+    def start_listen_sockets(self):
         ### All right, we're ready to accept customers:
-        for socktype, sock, info in sockets:
-            netlog("init_sockets(%s) will add %s socket %s (%s)", sockets, socktype, sock, info)
+        for socktype, sock, info, _ in self._socket_info:
+            netlog("init_sockets(%s) will add %s socket %s (%s)", self._socket_info, socktype, sock, info)
             self.socket_info[sock] = info
             self.idle_add(self.add_listen_socket, socktype, sock)
             if socktype=="unix-domain" and info:
@@ -667,6 +878,12 @@ class ServerCore(object):
     def touch_sockets(self):
         netlog("touch_sockets() unix socket paths=%s", self.unix_socket_paths)
         for sockpath in self.unix_socket_paths:
+            if not os.path.exists(sockpath):
+                if first_time("missing-socket-%s" % sockpath):
+                    log.warn("Warning: the unix domain socket cannot be found:")
+                    log.warn(" '%s'", sockpath)
+                    log.warn(" was it deleted by mistake?")
+                continue
             try:
                 os.utime(sockpath, None)
             except Exception:
@@ -691,7 +908,6 @@ class ServerCore(object):
         i = 1
         for key in packet_types:
             self._aliases[i] = key
-            self._reverse_aliases[key] = i
             i += 1
 
     def cleanup_udp_listeners(self):
@@ -717,102 +933,48 @@ class ServerCore(object):
     def add_listen_socket(self, socktype, sock):
         info = self.socket_info.get(sock)
         netlog("add_listen_socket(%s, %s) info=%s", socktype, sock, info)
-        try:
-            #ugly that we have different ways of starting sockets,
-            #TODO: abstract this into the socket class
-            self.socket_types[sock] = socktype
-            if socktype=="named-pipe":
-                #named pipe listener uses a thread:
-                sock.new_connection_cb = self._new_connection
-                sock.start()
-            elif socktype=="udp":
-                #socket_info = self.socket_info.get(sock)
-                from xpra.net.udp_protocol import UDPListener
-                udpl = UDPListener(sock, self.process_udp_packet)
-                self._udp_listeners.append(udpl)
-            else:
-                from xpra.gtk_common.gobject_compat import import_glib, is_gtk3
-                glib = import_glib()
-                sock.listen(5)
-                if is_gtk3():
-                    glib.io_add_watch(sock, glib.PRIORITY_DEFAULT, glib.IO_IN, self._new_connection, sock)
-                else:
-                    glib.io_add_watch(sock, glib.IO_IN, self._new_connection, sock, priority=glib.PRIORITY_DEFAULT)
-        except Exception as e:
-            netlog("add_listen_socket(%s, %s)", socktype, sock, exc_info=True)
-            netlog.error("Error: failed to listen on %s socket %s:", socktype, info or sock)
-            netlog.error(" %s", e)
+        cleanup = add_listen_socket(socktype, sock, info, self._new_connection, self._new_udp_connection)
+        if cleanup:
+            self.socket_cleanup.append(cleanup)
 
-    def _new_connection(self, listener, *args):
+    def _new_udp_connection(self, sock):
+        from xpra.net.udp_protocol import UDPListener
+        udpl = UDPListener(sock, self.process_udp_packet)
+        self._udp_listeners.append(udpl)
+
+    def _new_connection(self, socktype, listener, handle=0):
         """
             Accept the new connection,
             verify that there aren't too many,
             start a thread to dispatch it to the correct handler.
         """
+        log("_new_connection%s", (listener, socktype, handle))
         if self._closing:
             netlog("ignoring new connection during shutdown")
             return False
-        socktype = self.socket_types.get(listener)
         socket_info = self.socket_info.get(listener)
         assert socktype, "cannot find socket type for %s" % listener
         #TODO: just like add_listen_socket above, this needs refactoring
         if socktype=="named-pipe":
-            pipe_handle = args[0]
             from xpra.platform.win32.namedpipes.connection import NamedPipeConnection
-            conn = NamedPipeConnection(listener.pipe_name, pipe_handle)
+            conn = NamedPipeConnection(listener.pipe_name, handle)
             netlog.info("New %s connection received on %s", socktype, conn.target)
             return self.make_protocol(socktype, conn)
 
-        try:
-            sock, address = listener.accept()
-        except socket.error as e:
-            netlog("rejecting new connection on %s", listener, exc_info=True)
-            netlog.error("Error: cannot accept new connection:")
-            netlog.error(" %s", e)
+        conn = accept_connection(socktype, listener, self._socket_timeout)
+        if conn is None:
             return True
-        netlog("peer: %s", get_peercred(sock))
-        try:
-            peername = sock.getpeername()
-        except (OSError, IOError):
-            peername = address
         #limit number of concurrent network connections:
         if socktype not in ("unix-domain", ) and len(self._potential_protocols)>=self._max_connections:
             netlog.error("Error: too many connections (%i)", len(self._potential_protocols))
-            netlog.error(" ignoring new one: %s", peername)
-            sock.close()
+            netlog.error(" ignoring new one: %s", conn.endpoint)
+            conn.close()
             return True
-        sockname = sock.getsockname()
-        sock.settimeout(self._socket_timeout)
-        netlog("new_connection(%s) sock=%s, socket_info=%s, timeout=%s, address=%s, peername=%s. timeout=%s",
-               args, sock, socket_info, self._socket_timeout, address, peername, self._socket_timeout)
-        conn = SocketConnection(sock, sockname, address, peername, socktype)
-
         #from here on, we run in a thread, so we can poll (peek does)
-        start_thread(self.handle_new_connection, "new-%s-connection" % socktype, True, args=(conn, sock, address, socktype, peername, socket_info))
+        start_thread(self.handle_new_connection, "new-%s-connection" % socktype, True,
+                     args=(conn, socket_info))
         return True
 
-    def peek_connection(self, conn, timeout=1):
-        PEEK_SIZE = 8192
-        start = monotonic_time()
-        peek_data = b""
-        while not peek_data and monotonic_time()-start<timeout:
-            try:
-                peek_data = conn.peek(PEEK_SIZE)
-            except (OSError, IOError):
-                pass
-            except ValueError:
-                log("peek_connection(%s, %i) failed", conn, timeout, exc_info=True)
-                break
-            sleep(timeout/4.0)
-        line1 = b""
-        netlog("socket %s peek: got %i bytes", conn, len(peek_data))
-        if peek_data:
-            line1 = peek_data.splitlines()[0]
-            netlog("socket peek=%s", repr_ellipsized(peek_data, limit=512))
-            netlog("socket peek hex=%s", hexstr(peek_data[:128]))
-            netlog("socket peek line1=%s", repr_ellipsized(bytestostr(line1)))
-        return peek_data, line1
-
     def new_conn_err(self, conn, sock, socktype, socket_info, network_protocol,
                      msg="invalid packet format, not an xpra client?"):
         #not an xpra client
@@ -840,14 +1002,19 @@ class ServerCore(object):
     def force_close_connection(self, conn):
         try:
             conn.close()
-        except:
+        except (OSError, IOError):
             log("close_connection()", exc_info=True)
 
-    def handle_new_connection(self, conn, sock, address, socktype, peername, socket_info):
+    def handle_new_connection(self, conn, socket_info):
         """
             Use peek to decide what sort of connection this is,
             and start the appropriate handler for it.
         """
+        sock = conn._socket
+        address = conn.remote
+        socktype = conn.socktype
+        peername = conn.endpoint
+
         sockname = sock.getsockname()
         target = peername or sockname
         sock.settimeout(self._socket_timeout)
@@ -859,9 +1026,11 @@ class ServerCore(object):
         peek_data, line1 = None, None
         #rfb does not send any data, waits for a server packet
         if socktype!="rfb":
-            peek_data, line1 = self.peek_connection(conn, PEEK_TIMEOUT)
+            peek_data, line1 = peek_connection(conn)
 
         def ssl_wrap():
+            if not self._ssl_wrap_socket:
+                raise Exception("no ssl support")
             ssl_sock = self._ssl_wrap_socket(sock)
             ssllog("ssl wrapped socket(%s)=%s", sock, ssl_sock)
             if ssl_sock is None:
@@ -878,7 +1047,7 @@ class ServerCore(object):
                 packet_type = None
                 if peek_data[0] in ("P", ord("P")):
                     packet_type = "xpra"
-                elif line1.find("HTTP/")>0:
+                elif line1.find(b"HTTP/")>0:
                     packet_type = "HTTP"
                 if packet_type:
                     self.new_conn_err(conn, sock, socktype, socket_info, packet_type,
@@ -894,16 +1063,16 @@ class ServerCore(object):
                 http = True
             elif self.ssl_mode=="auto" or self.ssl_mode in TRUE_OPTIONS:
                 #look for HTTPS request to handle:
-                if line1.find("HTTP/")>0 or peek_data.find("\x08http/1.1")>0:
+                if line1.find(b"HTTP/")>0 or peek_data.find(b"\x08http/1.1")>0:
                     http = True
                 else:
                     ssl_conn.enable_peek()
-                    peek_data, line1 = self.peek_connection(ssl_conn, PEEK_TIMEOUT)
-                    http = line1.find("HTTP/")>0
+                    peek_data, line1 = peek_connection(ssl_conn)
+                    http = line1.find(b"HTTP/")>0
             if http and self._html:
                 self.start_http_socket(socktype, ssl_conn, True, peek_data)
             else:
-                log_new_connection(conn, socket_info)
+                log_new_connection(ssl_conn, socket_info)
                 self.make_protocol(socktype, ssl_conn)
             return
 
@@ -1001,7 +1170,7 @@ class ServerCore(object):
 
     def try_upgrade_to_rfb(self, proto):
         self.cancel_upgrade_to_rfb_timer(proto)
-        if proto._closed:
+        if proto.is_closed():
             return False
         conn = proto._conn
         netlog("may_upgrade_to_rfb() input_bytecount=%i", conn.input_bytecount)
@@ -1117,7 +1286,7 @@ class ServerCore(object):
                     http = True
                 else:
                     conn.enable_peek()
-                    peek_data, line1 = self.peek_connection(conn, PEEK_TIMEOUT)
+                    peek_data, line1 = peek_connection(conn)
                     http = line1.find(b"HTTP/")>0
             is_ssl = True
         else:
@@ -1186,11 +1355,12 @@ class ServerCore(object):
         #we start a new thread,
         #only so that the websocket handler thread is named correctly:
         start_thread(self.start_http, "%s-for-%s" % (tname, frominfo),
-                     daemon=True, args=(socktype, conn, is_ssl, req_info, conn.remote))
+                     daemon=True, args=(socktype, conn, is_ssl, req_info, line1, conn.remote))
 
-    def start_http(self, socktype, conn, is_ssl, req_info, frominfo):
-        httplog("start_http(%s, %s, %s, %s, %s) www dir=%s, headers dir=%s",
-                socktype, conn, is_ssl, req_info, frominfo, self._www_dir, self._http_headers_dir)
+    def start_http(self, socktype, conn, is_ssl, req_info, line1, frominfo):
+        httplog("start_http(%s, %s, %s, %s, %s, %s) www dir=%s, headers dir=%s",
+                socktype, conn, is_ssl, req_info, line1, frominfo,
+                self._www_dir, self._http_headers_dir)
         try:
             from xpra.net.websockets.handler import WebSocketRequestHandler
             sock = conn._socket
@@ -1214,6 +1384,7 @@ class ServerCore(object):
                 l = httplog.error
             l("Error: %s request failure", req_info)
             l(" for client %s:", pretty_socket(frominfo))
+            l(" request: '%s'", nonl(bytestostr(line1)))
             l(" %s", e)
         except Exception as e:
             wslog.error("Error: %s request failure for client %s:", req_info, pretty_socket(frominfo), exc_info=True)
@@ -1265,7 +1436,8 @@ class ServerCore(object):
             sock.settimeout(10)
             sock.connect((host, int(port)))
             sock.settimeout(None)
-            tcp_server_connection = SocketConnection(sock, sock.getsockname(), sock.getpeername(), "tcp-proxy-for-%s" % frominfo, "tcp")
+            tcp_server_connection = SocketConnection(sock, sock.getsockname(), sock.getpeername(),
+                                                     "tcp-proxy-for-%s" % frominfo, "tcp")
         except Exception as e:
             proxylog("start_tcp_proxy(%s, %s)", conn, frominfo, exc_info=True)
             proxylog.error("Error: failed to connect to TCP proxy endpoint: %s:%s", host, port)
@@ -1295,7 +1467,7 @@ class ServerCore(object):
 
     def is_timedout(self, protocol):
         #subclasses may override this method (ServerBase does)
-        v = not protocol._closed and protocol in self._potential_protocols and \
+        v = not protocol.is_closed() and protocol in self._potential_protocols and \
             protocol not in self._tcp_proxy_clients
         netlog("is_timedout(%s)=%s", protocol, v)
         return v
@@ -1324,7 +1496,7 @@ class ServerCore(object):
         netlog("send_disconnect(%s, %s)", proto, reasons)
         self.cancel_verify_connection_accepted(proto)
         self.cancel_upgrade_to_rfb_timer(proto)
-        if proto._closed:
+        if proto.is_closed():
             return
         proto.send_disconnect(reasons)
         self.timeout_add(1000, self.force_disconnect, proto)
@@ -1338,7 +1510,7 @@ class ServerCore(object):
 
     def disconnect_client(self, protocol, reason, *extra):
         netlog("disconnect_client(%s, %s, %s)", protocol, reason, extra)
-        if protocol and not protocol._closed:
+        if protocol and not protocol.is_closed():
             self.disconnect_protocol(protocol, reason, *extra)
 
     def disconnect_protocol(self, protocol, *reasons):
@@ -1346,10 +1518,17 @@ class ServerCore(object):
         i = str(reasons[0])
         if len(reasons)>1:
             i += " (%s)" % csv(reasons[1:])
+        proto_info = " %s" % protocol
         try:
-            proto_info = " %s" % protocol._conn.get_info().get("endpoint")
-        except:
-            proto_info = " %s" % protocol
+            conn = protocol._conn
+            info = conn.get_info()
+            endpoint = info.get("endpoint")
+            if endpoint:
+                proto_info = " %s" % (endpoint,)
+            else:
+                proto_info = " %s" % (conn.local,)
+        except (KeyError, AttributeError):
+            pass
         self._log_disconnect(protocol, "Disconnecting client%s:", proto_info)
         self._log_disconnect(protocol, " %s", i)
         self.cancel_verify_connection_accepted(protocol)
@@ -1381,7 +1560,7 @@ class ServerCore(object):
         self.cancel_verify_connection_accepted(proto)
         self.cancel_upgrade_to_rfb_timer(proto)
         if proto in self._potential_protocols:
-            if not proto._closed:
+            if not proto.is_closed():
                 self._log_disconnect(proto, "Connection lost")
             self._potential_protocols.remove(proto)
         #remove from UDP protocol map:
@@ -1592,7 +1771,7 @@ class ServerCore(object):
                 return
             #challenge has been sent already for this module
             if not challenge_response:
-                auth_failed("invalid state, challenge already sent - no response!")
+                auth_failed("invalid state, challenge already sent - no response found!")
                 return
             if not authenticator.authenticate(challenge_response, client_salt):
                 auth_failed("authentication failed")
@@ -1642,6 +1821,7 @@ class ServerCore(object):
                 raise Exception("Simulating a server error")
             self.hello_oked(proto, packet, c, auth_caps)
         except ClientException as e:
+            log("call_hello_oked(%s, %s, %s, %s)", proto, packet, repr_ellipsized(str(c)), auth_caps, exc_info=True)
             log.error("Error setting up new connection for")
             log.error(" %s:", proto)
             log.error(" %s", e)
@@ -1677,8 +1857,7 @@ class ServerCore(object):
     def accept_client(self, proto, c):
         #max packet size from client (the biggest we can get are clipboard packets)
         netlog("accept_client(%s, %s)", proto, c)
-        #TODO: when we add the code to upload files,
-        #this will need to be increased up to file-size-limit
+        #note: when uploading files, we send them in chunks smaller than this size
         proto.max_packet_size = 1024*1024  #1MB
         proto.parse_remote_caps(c)
         self.accept_protocol(proto)
@@ -1718,6 +1897,7 @@ class ServerCore(object):
                         "elapsed_time"          : int(now - self.start_time),
                         "server_type"           : "core",
                         "server.mode"           : self.get_server_mode(),
+                        "hostname"              : socket.gethostname(),
                         })
         if source is None or source.wants_features:
             capabilities.update({
@@ -1797,12 +1977,15 @@ class ServerCore(object):
         return get_thread_info(proto)
 
     def get_minimal_server_info(self):
+        now = time()
         info = {
             "mode"              : self.get_server_mode(),
             "session-type"      : self.session_type,
             "type"              : "Python",
             "python"            : {"version" : python_platform.python_version()},
             "start_time"        : int(self.start_time),
+            "current_time"      : int(now),
+            "elapsed_time"      : int(now - self.start_time),
             "uuid"              : self.uuid,
             }
         return info
@@ -1819,6 +2002,11 @@ class ServerCore(object):
             "executable"        : sys.executable,
             "idle-timeout"      : int(self.server_idle_timeout),
             })
+        if self.pidfile:
+            si["pidfile"] = {
+                "path"  : self.pidfile,
+                "inode" : self.pidinode,
+                }
         if POSIX:
             si["load"] = tuple(int(x*1000) for x in os.getloadavg())
         if self.original_desktop_display:
@@ -1851,7 +2039,8 @@ class ServerCore(object):
                        ""                   : self._html,
                        "dir"                : self._www_dir or "",
                        "http-headers-dir"   : self._http_headers_dir or "",
-                       }
+                       },
+                   "mdns"           : self.mdns,
                    })
         up("network", ni)
         up("threads",   self.get_thread_info(proto))
@@ -1871,7 +2060,7 @@ class ServerCore(object):
 
     def get_socket_info(self):
         si = {}
-        for socktype, _, info in self._socket_info:
+        for socktype, _, info, _ in self._socket_info:
             if info:
                 si.setdefault(socktype, {}).setdefault("listeners", []).append(info)
         for socktype, auth_classes in self.auth_classes.items():
@@ -1889,6 +2078,7 @@ class ServerCore(object):
         handler = None
         try:
             packet_type = bytestostr(packet[0])
+            may_log_packet(packet_type, packet)
             handler = self._default_packet_handlers.get(packet_type)
             if handler:
                 netlog("process packet %s", packet_type)
